#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;

namespace HeuristicLab.Optimization.Networks {
  [Item("AlgorithmNode", "A node of an optimization network which contains a HeuristicLab algorithm.")]
  [StorableClass]
  public class AlgorithmNode : Node, IAlgorithmNode {
    new public PortCollection Ports {
      get { return base.Ports; }
    }

    [Storable]
    private IAlgorithm algorithm;
    public IAlgorithm Algorithm {
      get { return algorithm; }
      set {
        if (value != algorithm) {
          DeregisterAlgorithmEvents();
          algorithm = value;
          RegisterAlgorithmEvents();
          OnAlgorithmChanged();
        }
      }
    }

    [StorableConstructor]
    protected AlgorithmNode(bool deserializing) : base(deserializing) { }
    protected AlgorithmNode(AlgorithmNode original, Cloner cloner)
      : base(original, cloner) {
      algorithm = cloner.Clone(original.algorithm);
      RegisterPortsEvents();
      RegisterAlgorithmEvents();
    }
    public AlgorithmNode()
      : base("AlgorithmNode") {
      RegisterPortsEvents();
    }
    public AlgorithmNode(string name)
      : base(name) {
      RegisterPortsEvents();
    }
    public AlgorithmNode(string name, string description)
      : base(name, description) {
      RegisterPortsEvents();
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterPortsEvents();
      RegisterAlgorithmEvents();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new AlgorithmNode(this, cloner);
    }

    private void UpdateParameter(IInputPort port) {
      if (algorithm != null) {
        IParameter param = null;
        if (algorithm.Parameters.TryGetValue(port.Name, out param)) {
          IValueParameter p = param as IValueParameter;
          if ((p != null) && (port.Value != null)) {
            p.Value = (IItem)port.Value;
          }
        }
      }
    }
    private void UpdateOutputPort(IResult result) {
      IPort port = null;
      if (Ports.TryGetValue(result.Name, out port)) {
        IOutputPort p = port as IOutputPort;
        if (p != null) {
          p.Value = result.Value;
        }
      }
    }
    private void UpdateOutputPort(IOutputPort port) {
      if ((algorithm != null) && (algorithm.Results != null)) {
        IResult result = null;
        if (algorithm.Results.TryGetValue(port.Name, out result)) {
          port.Value = result != null ? result.Value : null;
        } else {
          port.Value = null;
        }
      }
    }

    public event EventHandler AlgorithmChanged;
    protected virtual void OnAlgorithmChanged() {
      var handler = AlgorithmChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    #region Ports Events
    protected override void RegisterPortsEvents() {
      base.RegisterPortsEvents();
      foreach (var p in Ports)
        RegisterPortEvents(p);
    }
    protected override void Ports_ItemsAdded(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_ItemsAdded(sender, e);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    protected override void Ports_ItemsRemoved(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_ItemsRemoved(sender, e);
      foreach (var p in e.Items)
        DeregisterPortEvents(p);
    }
    protected override void Ports_ItemsReplaced(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_ItemsReplaced(sender, e);
      foreach (var p in e.OldItems)
        DeregisterPortEvents(p);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    protected override void Ports_CollectionReset(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_CollectionReset(sender, e);
      foreach (var p in e.OldItems)
        DeregisterPortEvents(p);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    #endregion

    #region Port Events
    private void RegisterPortEvents(IPort port) {
      IInputPort i = port as IInputPort;
      if (i != null) {
        i.NameChanged += InputPort_NameChanged;
        i.ValueChanged += InputPort_ValueChanged;
        UpdateParameter(i);
      }
      IOutputPort o = port as IOutputPort;
      if (o != null) {
        o.NameChanged += OutputPort_NameChanged;
        UpdateOutputPort(o);
      }
    }
    private void DeregisterPortEvents(IPort port) {
      IInputPort i = port as IInputPort;
      if (i != null) {
        i.NameChanged -= InputPort_NameChanged;
        i.ValueChanged -= InputPort_ValueChanged;
      }
      IOutputPort o = port as IOutputPort;
      if (o != null) {
        o.NameChanged -= OutputPort_NameChanged;
      }
    }
    private void InputPort_NameChanged(object sender, EventArgs e) {
      UpdateParameter((IInputPort)sender);
    }
    private void InputPort_ValueChanged(object sender, EventArgs e) {
      UpdateParameter((IInputPort)sender);
    }
    private void OutputPort_NameChanged(object sender, EventArgs e) {
      UpdateOutputPort((IOutputPort)sender);
    }
    #endregion

    #region Algorithm Events
    private void RegisterAlgorithmEvents() {
      if (algorithm != null) {
        algorithm.Prepared += Algorithm_Prepared;
        RegisterResultsEvents();
      }
    }

    private void DeregisterAlgorithmEvents() {
      if (algorithm != null) {
        algorithm.Prepared -= Algorithm_Prepared;
        DeregisterResultsEvents();
      }
    }
    private void Algorithm_Prepared(object sender, EventArgs e) {
      RegisterResultsEvents();
    }
    #endregion

    #region Results Events
    private void RegisterResultsEvents() {
      if ((algorithm != null) && (algorithm.Results != null)) {
        algorithm.Results.ItemsAdded += Results_ItemsAdded;
        algorithm.Results.ItemsRemoved += Results_ItemsRemoved;
        algorithm.Results.ItemsReplaced += Results_ItemsReplaced;
        algorithm.Results.CollectionReset += Results_CollectionReset;
        foreach (var r in algorithm.Results)
          RegisterResultEvents(r);
      }
    }
    private void DeregisterResultsEvents() {
      if ((algorithm != null) && (algorithm.Results != null)) {
        algorithm.Results.ItemsAdded -= Results_ItemsAdded;
        algorithm.Results.ItemsRemoved -= Results_ItemsRemoved;
        algorithm.Results.ItemsReplaced -= Results_ItemsReplaced;
        algorithm.Results.CollectionReset -= Results_CollectionReset;
        foreach (var r in algorithm.Results)
          DeregisterResultEvents(r);
      }
    }
    private void Results_ItemsAdded(object sender, Collections.CollectionItemsChangedEventArgs<IResult> e) {
      foreach (var r in e.Items)
        RegisterResultEvents(r);
    }
    private void Results_ItemsRemoved(object sender, Collections.CollectionItemsChangedEventArgs<IResult> e) {
      foreach (var r in e.Items)
        DeregisterResultEvents(r);
    }
    private void Results_ItemsReplaced(object sender, Collections.CollectionItemsChangedEventArgs<IResult> e) {
      foreach (var r in e.OldItems)
        DeregisterResultEvents(r);
      foreach (var r in e.Items)
        RegisterResultEvents(r);
    }
    private void Results_CollectionReset(object sender, Collections.CollectionItemsChangedEventArgs<IResult> e) {
      foreach (var r in e.OldItems)
        DeregisterResultEvents(r);
      foreach (var r in e.Items)
        RegisterResultEvents(r);
    }
    #endregion

    #region Result Events
    private void RegisterResultEvents(IResult result) {
      if (result != null) {
        result.ValueChanged += Result_ValueChanged;
        result.ToStringChanged += Result_ToStringChanged;
        UpdateOutputPort(result);
      }
    }
    private void DeregisterResultEvents(IResult result) {
      if (result != null) {
        result.ValueChanged -= Result_ValueChanged;
        result.ToStringChanged += Result_ToStringChanged;
      }
    }
    private void Result_ValueChanged(object sender, EventArgs e) {
      UpdateOutputPort((IResult)sender);
    }
    private void Result_ToStringChanged(object sender, EventArgs e) {
      UpdateOutputPort((IResult)sender);
    }
    #endregion
  }
}
